"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const protobufjs_1 = require("protobufjs");
const uuid_1 = require("uuid");
const bplist = require("bplist-creator");
const watson_developer_cloud_1 = require("watson-developer-cloud");
const opus = require("node-opus");
const connection_1 = require("./connection");
const pairing_1 = require("./pairing");
const verifier_1 = require("./verifier");
const now_playing_info_1 = require("./now-playing-info");
const supported_command_1 = require("./supported-command");
const typed_events_1 = require("./typed-events");
const message_1 = require("./message");
const number_1 = require("./util/number");
class AppleTV extends typed_events_1.default {
    constructor(service) {
        super();
        this.service = service;
        this.pairingId = uuid_1.v4();
        this.service = service;
        this.name = service.txtRecord.Name;
        if (service.addresses.length > 1) {
            this.address = service.addresses[1];
        }
        else {
            this.address = service.addresses[0];
        }
        this.port = service.port;
        this.uid = service.txtRecord.UniqueIdentifier;
        this.connection = new connection_1.Connection(this);
        let that = this;
        this.connection.on('message', (message) => {
            that.emit('message', message);
            if (message.type == message_1.Message.Type.SetStateMessage) {
                if (message.payload == null) {
                    that.emit('nowPlaying', null);
                    return;
                }
                if (message.payload.nowPlayingInfo) {
                    let info = new now_playing_info_1.NowPlayingInfo(message.payload);
                    that.emit('nowPlaying', info);
                }
                if (message.payload.supportedCommands) {
                    let commands = (message.payload.supportedCommands.supportedCommands || [])
                        .map(sc => {
                        return new supported_command_1.SupportedCommand(sc.command, sc.enabled || false, sc.canScrub || false);
                    });
                    that.emit('supportedCommands', commands);
                }
                if (message.payload.playbackQueue) {
                    that.emit('playbackQueue', message.payload.playbackQueue);
                }
            }
            else if (message.type == message_1.Message.Type.SetRecordingStateMessage) {
                that.isRecording = message.payload.state == 1;
                that.emit('recordingState', message.payload.state == 1);
            }
        })
            .on('connect', () => {
            that.emit('connect');
        })
            .on('close', () => {
            that.emit('close');
        })
            .on('error', (error) => {
            that.emit('error', error);
        })
            .on('debug', (message) => {
            that.emit('debug', message);
        });
        var queuePollTimer = null;
        this._on('newListener', (event, listener) => {
            if (queuePollTimer == null && (event == 'nowPlaying' || event == 'supportedCommands' || event == 'recordingState')) {
                queuePollTimer = setInterval(() => {
                    if (that.connection.isOpen) {
                        that.requestPlaybackQueueWithWait({
                            length: 1,
                            location: 0
                        }, false).then(() => { }).catch(error => { });
                    }
                }, 2000);
            }
        });
        this._on('removeListener', (event, listener) => {
            if (queuePollTimer != null && (event == 'nowPlaying' || event == 'supportedCommands' || event == 'recordingState')) {
                let listenerCount = that.listenerCount('nowPlaying') + that.listenerCount('supportedCommands') + that.listenerCount('recordingState');
                if (listenerCount == 0) {
                    clearInterval(queuePollTimer);
                    queuePollTimer = null;
                }
            }
        });
    }
    /**
    * Pair with an already discovered AppleTV.
    * @returns A promise that resolves to the AppleTV object.
    */
    pair() {
        let pairing = new pairing_1.Pairing(this);
        return pairing.initiatePair();
    }
    /**
    * Opens a connection to the AppleTV over the MRP protocol.
    * @param credentials  The credentials object for this AppleTV
    * @returns A promise that resolves to the AppleTV object.
    */
    openConnection(credentials) {
        let that = this;
        if (credentials) {
            this.pairingId = credentials.pairingId;
        }
        return this.connection
            .open()
            .then(() => {
            return that.sendIntroduction();
        })
            .then(() => {
            that.credentials = credentials;
            if (credentials) {
                let verifier = new verifier_1.Verifier(that);
                return verifier.verify()
                    .then(keys => {
                    that.credentials.readKey = keys['readKey'];
                    that.credentials.writeKey = keys['writeKey'];
                    that.emit('debug', "DEBUG: Keys Read=" + that.credentials.readKey.toString('hex') + ", Write=" + that.credentials.writeKey.toString('hex'));
                    return that.sendConnectionState();
                });
            }
            else {
                return null;
            }
        })
            .then(() => {
            if (credentials) {
                return that.sendClientUpdatesConfig();
            }
            else {
                return null;
            }
        })
            .then(() => {
            return Promise.resolve(that);
        });
    }
    /**
    * Closes the connection to the Apple TV.
    */
    closeConnection() {
        this.connection.close();
    }
    /**
    * Send a Protobuf message to the AppleTV. This is for advanced usage only.
    * @param definitionFilename  The Protobuf filename of the message type.
    * @param messageType  The name of the message.
    * @param body  The message body
    * @param waitForResponse  Whether or not to wait for a response before resolving the Promise.
    * @returns A promise that resolves to the response from the AppleTV.
    */
    sendMessage(definitionFilename, messageType, body, waitForResponse, priority = 0) {
        return protobufjs_1.load(path.resolve(__dirname + "/protos/" + definitionFilename + ".proto"))
            .then(root => {
            let type = root.lookupType(messageType);
            return type.create(body);
        })
            .then(message => {
            return this.connection
                .send(message, waitForResponse, priority, this.credentials);
        });
    }
    /**
    * Registers a voice input device to send Siri commands
    * @returns A Promise that resolves to the device id returned.
    */
    registerVoiceInputDevice() {
        let formatSettingsData = bplist({
            AVFormatIDKey: 1869641075,
            AVNumberOfChannelsKey: 1,
            AVSampleRateKey: new bplist.Real(16000)
        });
        let that = this;
        return this.sendMessage('RegisterVoiceInputDeviceMessage', 'RegisterVoiceInputDeviceMessage', {
            deviceDescriptor: {
                defaultFormat: {
                    formatSettingsPlistData: formatSettingsData,
                },
                supportedFormats: [{
                        formatSettingsPlistData: formatSettingsData,
                    }]
            }
        }, false)
            .then(message => {
            return that;
        });
    }
    sendVoiceInput() {
        let that = this;
        let formatSettingsData = bplist({
            AVFormatIDKey: 1869641075,
            AVNumberOfChannelsKey: 1,
            AVSampleRateKey: new bplist.Real(16000)
        });
        return this.sendKeyCommand(AppleTV.Key.Record)
            .then(device => {
            return new Promise((resolve, reject) => {
                let listener = (isRecording) => {
                    if (isRecording) {
                        resolve(device);
                        device.removeListener('recordingState', listener);
                    }
                };
                device.on('recordingState', listener);
                setTimeout(() => {
                    reject(new Error("Timed out waiting for recording state."));
                    device.removeListener('recordingState', listener);
                }, 5000);
            });
        })
            .then(device => {
            let tts = new watson_developer_cloud_1.TextToSpeechV1({
                username: '63dd2f3d-be9a-4f6c-9f4a-0e8d90f5ec8b',
                password: 'FlljeTfHyFgI',
                version: 'v1'
            });
            let params = {
                text: 'open plex',
                accept: 'audio/ogg; codec=opus'
            };
            return new Promise((resolve, reject) => {
                tts.synthesize(params, (error, body, response) => {
                    if (error) {
                        reject(error);
                    }
                    else {
                        resolve(body);
                    }
                });
            })
                .then(buffer => {
                let encoder = new opus.OpusEncoder(16000);
                let decoded = encoder.decode(buffer, 160);
                console.log(decoded);
                let maximumPacketSize = 1024;
                let packetCount = Math.ceil(buffer.byteLength / maximumPacketSize);
                // var currentIndex = 0;
                // let packectDescriptions = Array(packetCount)
                //   .fill(0)
                //   .map(() => {
                //     return {
                //       dataByteSize: 41,
                //       startOffset: 0,
                //       variableFramesInPacket: 0,
                //     };
                //   });
                return device.sendMessage('SendVoiceInputMessage', 'SendVoiceInputMessage', {
                    dataBlock: {
                        buffer: {
                            formatSettings: {
                                formatSettingsPlistData: formatSettingsData,
                            },
                            contents: buffer,
                            maximumPacketSize: maximumPacketSize,
                            packetCapacity: packetCount,
                            packetCount: packetCount,
                            packetDescriptions: [{
                                    dataByteSize: 41,
                                    startOffset: 0,
                                    variableFramesInPacket: 0,
                                }, {
                                    dataByteSize: 39,
                                    startOffset: 41,
                                    variableFramesInPacket: 0,
                                }, {
                                    dataByteSize: 43,
                                    startOffset: 80,
                                    variableFramesInPacket: 0,
                                }, {
                                    dataByteSize: 38,
                                    startOffset: 123,
                                    variableFramesInPacket: 0,
                                }, {
                                    dataByteSize: 39,
                                    startOffset: 161,
                                    variableFramesInPacket: 0,
                                }, {
                                    dataByteSize: 42,
                                    startOffset: 200,
                                    variableFramesInPacket: 0,
                                }, {
                                    dataByteSize: 43,
                                    startOffset: 242,
                                    variableFramesInPacket: 0,
                                }],
                            gain: 0.05468025803565979,
                            time: {
                                sampleRate: 16000,
                                timestamp: 176.9877464166667,
                            }
                        }
                    }
                }, false, 1);
            });
        })
            .then(message => {
            return that.sendKeyCommand(AppleTV.Key.Record);
        })
            .then(device => {
            return new Promise((resolve, reject) => {
                let listener = (isRecording) => {
                    if (!isRecording) {
                        resolve(device);
                        device.removeListener('recordingState', listener);
                    }
                };
                device.on('recordingState', listener);
                setTimeout(() => {
                    reject(new Error("Timed out waiting for recording state."));
                    device.removeListener('recordingState', listener);
                }, 5000);
            });
        });
    }
    /**
    * Requests the current playback queue from the Apple TV.
    * @param options Options to send
    * @returns A Promise that resolves to a NewPlayingInfo object.
    */
    requestPlaybackQueue(options) {
        return this.requestPlaybackQueueWithWait(options, true);
    }
    /**
    * Send a key command to the AppleTV.
    * @param key The key to press.
    * @returns A promise that resolves to the AppleTV object after the message has been sent.
    */
    sendKeyCommand(key) {
        switch (key) {
            case AppleTV.Key.Up:
                return this.sendKeyPressAndRelease(1, 0x8C);
            case AppleTV.Key.Down:
                return this.sendKeyPressAndRelease(1, 0x8D);
            case AppleTV.Key.Left:
                return this.sendKeyPressAndRelease(1, 0x8B);
            case AppleTV.Key.Right:
                return this.sendKeyPressAndRelease(1, 0x8A);
            case AppleTV.Key.Menu:
                return this.sendKeyPressAndRelease(1, 0x86);
            case AppleTV.Key.Play:
                return this.sendKeyPressAndRelease(12, 0xB0);
            case AppleTV.Key.Pause:
                return this.sendKeyPressAndRelease(12, 0xB1);
            case AppleTV.Key.Next:
                return this.sendKeyPressAndRelease(12, 0xB5);
            case AppleTV.Key.Previous:
                return this.sendKeyPressAndRelease(12, 0xB6);
            case AppleTV.Key.Suspend:
                return this.sendKeyPressAndRelease(1, 0x82);
            case AppleTV.Key.Record:
                return this.sendKeyPress(12, 4, !this.isRecording);
        }
    }
    sendKeyPressAndRelease(usePage, usage) {
        let that = this;
        return this.sendKeyPress(usePage, usage, true)
            .then(() => {
            return that.sendKeyPress(usePage, usage, false);
        });
    }
    sendKeyPress(usePage, usage, down) {
        let time = Buffer.from('438922cf08020000', 'hex');
        let data = Buffer.concat([
            number_1.default.UInt16toBufferBE(usePage),
            number_1.default.UInt16toBufferBE(usage),
            down ? number_1.default.UInt16toBufferBE(1) : number_1.default.UInt16toBufferBE(0)
        ]);
        let body = {
            hidEventData: Buffer.concat([
                time,
                Buffer.from('00000000000000000100000000000000020' + '00000200000000300000001000000000000', 'hex'),
                data,
                Buffer.from('0000000000000001000000', 'hex')
            ])
        };
        let that = this;
        return this.sendMessage("SendHIDEventMessage", "SendHIDEventMessage", body, false)
            .then(() => {
            return that;
        });
    }
    requestPlaybackQueueWithWait(options, waitForResponse) {
        var params = options;
        params.requestID = uuid_1.v4();
        if (options.artworkSize) {
            params.artworkWidth = options.artworkSize.width;
            params.artworkHeight = options.artworkSize.height;
            delete params.artworkSize;
        }
        return this.sendMessage("PlaybackQueueRequestMessage", "PlaybackQueueRequestMessage", params, waitForResponse);
    }
    sendIntroduction() {
        let body = {
            uniqueIdentifier: this.pairingId,
            name: 'node-appletv',
            localizedModelName: 'iPhone',
            systemBuildVersion: '14G60',
            applicationBundleIdentifier: 'com.apple.TVRemote',
            applicationBundleVersion: '320.15',
            protocolVersion: 1,
            allowsPairing: true,
            lastSupportedMessageType: 45,
            supportsSystemPairing: true,
        };
        return this.sendMessage('DeviceInfoMessage', 'DeviceInfoMessage', body, true);
    }
    sendConnectionState() {
        let that = this;
        return protobufjs_1.load(path.resolve(__dirname + "/protos/SetConnectionStateMessage.proto"))
            .then(root => {
            let type = root.lookupType('SetConnectionStateMessage');
            let stateEnum = type.lookupEnum('ConnectionState');
            let message = type.create({
                state: stateEnum.values['Connected']
            });
            return that
                .connection
                .send(message, false, 0, that.credentials);
        });
    }
    sendClientUpdatesConfig() {
        let message = {
            artworkUpdates: true,
            nowPlayingUpdates: true,
            volumeUpdates: true,
            keyboardUpdates: true
        };
        return this.sendMessage('ClientUpdatesConfigMessage', 'ClientUpdatesConfigMessage', message, false);
    }
    sendWakeDevice() {
        return this.sendMessage('WakeDeviceMessage', 'WakeDeviceMessage', {}, false);
    }
}
exports.AppleTV = AppleTV;
(function (AppleTV) {
    /** An enumeration of key presses available.
    */
    let Key;
    (function (Key) {
        Key[Key["Up"] = 0] = "Up";
        Key[Key["Down"] = 1] = "Down";
        Key[Key["Left"] = 2] = "Left";
        Key[Key["Right"] = 3] = "Right";
        Key[Key["Menu"] = 4] = "Menu";
        Key[Key["Play"] = 5] = "Play";
        Key[Key["Pause"] = 6] = "Pause";
        Key[Key["Next"] = 7] = "Next";
        Key[Key["Previous"] = 8] = "Previous";
        Key[Key["Suspend"] = 9] = "Suspend";
        Key[Key["Record"] = 10] = "Record";
    })(Key = AppleTV.Key || (AppleTV.Key = {}));
    /** Convert a string representation of a key to the correct enum type.
    * @param string  The string.
    * @returns The key enum value.
    */
    function key(string) {
        if (string == "up") {
            return AppleTV.Key.Up;
        }
        else if (string == "down") {
            return AppleTV.Key.Down;
        }
        else if (string == "left") {
            return AppleTV.Key.Left;
        }
        else if (string == "right") {
            return AppleTV.Key.Right;
        }
        else if (string == "menu") {
            return AppleTV.Key.Menu;
        }
        else if (string == "play") {
            return AppleTV.Key.Play;
        }
        else if (string == "pause") {
            return AppleTV.Key.Pause;
        }
        else if (string == "next") {
            return AppleTV.Key.Next;
        }
        else if (string == "previous") {
            return AppleTV.Key.Previous;
        }
        else if (string == "suspend") {
            return AppleTV.Key.Suspend;
        }
        else if (string == "record") {
            return AppleTV.Key.Record;
        }
    }
    AppleTV.key = key;
})(AppleTV = exports.AppleTV || (exports.AppleTV = {}));
